#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <curses.h>
#include <signal.h>
#include <locale.h>
#include <termios.h>
#include <wchar.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netdb.h>

#define PROTVER			18

#define CS_NO			0
#define CS_LOBBY		1
#define CS_PLAYING		2
#define CS_MISSED		3

#define MAXCLIENTS		64
#define MSGLENGTH		128

#define EP_ANIMATE		0
#define EP_ANIMATEINPUT		1
#define EP_VOTE			2

int status = CS_LOBBY;
int evaluating = 0;
int eval_phase;
int sock;
volatile int got_ctrl_c = 0;
char serverid[MSGLENGTH];
char *nick;
wchar_t inpbuf[17];
int inppos = 0;
char *gameovertxt = 0;
char animate_path[16];
int animate_len, animate_pos, animate_flag = 0;
int votecursor;
int gametime = 180;
int dispheight, dispwidth;

struct contrib {
	wchar_t		*word;
	int		flashthresh;
};

struct client {
	int		status;
	char		name[MSGLENGTH];
	struct contrib	*contrib;
	int		ncontrib;
	int		ncontriballoc;
} client[MAXCLIENTS];

struct game {
	wchar_t		box[16];
	int		nsol;
	int		nsol_left;
	int		timer;
} game;

struct suggest {
	wchar_t		word[32];
	int		vote;
} *suggest = 0;

int nsuggest = 0, nsuggestalloc = 0;

volatile int got_winch = 0;

void sighandler(int sig) {
	switch(sig) {
		case SIGINT:
			got_ctrl_c = 1;
			break;
		case SIGALRM:
			if(game.timer) game.timer--;
			if(animate_flag) {
				animate_pos++;
				if(animate_pos > animate_len * 3 + 12) animate_flag = 0;
			}
			signal(SIGALRM, sighandler);
			break;
		case SIGWINCH:
			got_winch = 1;
			signal(SIGWINCH, sighandler);
			break;
	}
}

void msg(char *msg, ...) {
	char buf[MSGLENGTH+1];
	va_list va;

	va_start(va, msg);
	memset(buf, 0, sizeof(buf));
	vsnprintf(buf, sizeof(buf), msg, va);
	va_end(va);
	write(sock, buf, MSGLENGTH);
}

void disp_lobby() {
	char buf[1024];
	int i, y, g = 0;

	mvaddstr(2, 0, "Currently in the lobby:");
	y = 4;
	for(i = 0; i < MAXCLIENTS; i++) {
		if(client[i].status == CS_LOBBY) {
			mvaddstr(y++, 8, client[i].name);
		}
	}
	if(y > 4) y++;

	for(i = 0; i < MAXCLIENTS; i++) {
		if(client[i].status == CS_PLAYING) {
			g = 1;
		}
	}
	if(g) {
		if(evaluating) {
			mvaddstr(y++, 0, "Currently evaluating:");
		} else {
			snprintf(buf, sizeof(buf), "Currently playing (%d:%02d left):",
				(game.timer + 9) / 10 / 60,
				((game.timer + 9) / 10) % 60);
			mvaddstr(y++, 0, buf);
		}
		y++;
		for(i = 0; i < MAXCLIENTS; i++) {
			if(client[i].status == CS_PLAYING) {
				mvaddstr(y++, 8, client[i].name);
			}
		}
	} else {
		mvaddstr(y++, 0, "(Press [SPACE] to start a new game)");
		y++;
	}
	move(y, 0);
}

void disp_playing() {
	int x, y, i, j, k;
	int firstx, xpos, ypos, enterx, entery;
	int maxlen;
	char buf[1024];
	char boardgfx[9*17] = 
		".---t---t---t---,"
		"|   |   |   |   |"
		"l---+---+---+---r"
		"|   |   |   |   |"
		"l---+---+---+---r"
		"|   |   |   |   |"
		"l---+---+---+---r"
		"|   |   |   |   |"
		"`---b---b---b---'";

	for(y = 0; y < 9; y++) {
		for(x = 0; x < 17; x++) {
			switch(boardgfx[y*17+x]) {
				case '+': mvaddch(2 + y, x, ACS_PLUS); break;
				case '-': mvaddch(2 + y, x, ACS_HLINE); break;
				case '|': mvaddch(2 + y, x, ACS_VLINE); break;
				case '.': mvaddch(2 + y, x, ACS_ULCORNER); break;
				case ',': mvaddch(2 + y, x, ACS_URCORNER); break;
				case '`': mvaddch(2 + y, x, ACS_LLCORNER); break;
				case '\'': mvaddch(2 + y, x, ACS_LRCORNER); break;
				case 'b': mvaddch(2 + y, x, ACS_BTEE); break;
				case 't': mvaddch(2 + y, x, ACS_TTEE); break;
				case 'l': mvaddch(2 + y, x, ACS_LTEE); break;
				case 'r': mvaddch(2 + y, x, ACS_RTEE); break;
			}
		}
	}
	attron(A_BOLD);
	for(y = 0; y < 4; y++) {
		for(x = 0; x < 4; x++) {
			color_set(1, 0);
			if(animate_flag) {
				for(i = 0; i <= (animate_pos / 3) && i < animate_len; i++) {
					if(y * 4 + x == animate_path[i]) {
						color_set(3, 0);
					}
				}
			}
			k = wctomb(buf, game.box[y * 4 + x]);
			if(k >= 0) {
				buf[k] = 0;
				mvaddch(3 + y * 2, 1 + x * 4, ' ');
				addstr(buf);
				if(k < 2) addch(' ');
			}
		}
	}
	attroff(A_BOLD);
	color_set(0, 0);

	if(evaluating) {
		mvaddstr(3, 24, gameovertxt);
		snprintf(buf, sizeof(buf), "%d of %d words missed (%d%% solved)", 
			 game.nsol_left, game.nsol, 
			 ((game.nsol-game.nsol_left)*100)/game.nsol);
		mvaddstr(8, 24, buf);
		if (game.nsol != game.nsol_left) {
			snprintf(buf, sizeof(buf), "You solved %d of %d (%d%%)", 
				 client[0].ncontrib, game.nsol-game.nsol_left, 
				 (client[0].ncontrib*100)/(game.nsol-game.nsol_left));
			mvaddstr(9, 24, buf);
		}
		switch(eval_phase) {
			case EP_ANIMATE:
				mvaddstr(5, 24, "[A]     Animate a word");
				mvaddstr(6, 24, "[ENTER] Continue...");
				
				break;
			case EP_ANIMATEINPUT:
				mvaddstr(5, 24, "Animate: ");
				break;
			case EP_VOTE:
				attron(A_BOLD);
				mvaddstr(3, 24, "Add any of these words to the dictionary?");
				attroff(A_BOLD);
				mvaddstr(5, 24, "[UP/DOWN] Select word");
				mvaddstr(6, 24, "[ENTER]   Vote for selected word");
				break;
		}
	} else {
		snprintf(buf, sizeof(buf), "%d:%02d", (game.timer + 9) / 10 / 60, ((game.timer + 9) / 10) % 60);
		mvaddstr(3, 24, buf);

		snprintf(buf, sizeof(buf), "%d of %d words left (%d%% solved)", 
			 game.nsol_left, game.nsol, 
			 ((game.nsol-game.nsol_left)*100)/game.nsol);
		mvaddstr(5, 24, buf);
		if (game.nsol != game.nsol_left) {
			snprintf(buf, sizeof(buf), "You solved %d of %d (%d%%)", 
				 client[0].ncontrib, game.nsol-game.nsol_left, 
				 (client[0].ncontrib*100)/(game.nsol-game.nsol_left));
			mvaddstr(6, 24, buf);
		}
	}

	if(evaluating && eval_phase == EP_VOTE) {
		ypos = 12;
		xpos = 4;
		y = 0;
		for(i = 0; i <= nsuggest; i++) {
			if(i == votecursor) attron(A_REVERSE);
			if(i == nsuggest) {
				mvaddstr(ypos + y + 1, xpos, "Done");
			} else {
				if(suggest[i].vote) {
					mvaddstr(ypos + y, xpos, "[X] ");
				} else {
					mvaddstr(ypos + y, xpos, "[ ] ");
				}
				if(wcstombs(buf, suggest[i].word, sizeof(buf)) >= 0) {
					addstr(buf);
				}
			}
			getyx(stdscr, k, j);
			for(x = j; x < xpos + 22; x++) addch(' ');
			attroff(A_REVERSE);
			y++;
			if(ypos + y >= dispheight - 1) {
				y = 0;
				xpos += 24;
			}
		}
	} else {
		xpos = 0;
		ypos = 12;
		for(i = 0; i < MAXCLIENTS; i++) {
			if(client[i].status == CS_PLAYING
			|| (evaluating && client[i].status == CS_MISSED && client[i].ncontrib)) {
				y = 0;
				firstx = xpos;
				if(client[i].status == CS_MISSED) color_set(2, 0);
				attron(A_BOLD);
				mvaddstr(ypos + y, xpos, client[i].name);
				maxlen = strlen(client[i].name);
				if(!i) maxlen = 16;
				attroff(A_BOLD);
				y += 2;
				for(j = 0; j < client[i].ncontrib; j++) {
					if(wcstombs(buf, client[i].contrib[j].word, sizeof(buf)) > 0) {
						if(wcslen(client[i].contrib[j].word) > maxlen) {
							maxlen = wcslen(client[i].contrib[j].word);
						}
						if(client[i].contrib[j].flashthresh < game.timer) {
							attron(A_REVERSE);
						}
						mvaddstr(ypos + y, xpos, buf);
						attroff(A_REVERSE);
						y++;
						if(ypos + y >= dispheight - 1) {
							y = 2;
							xpos += maxlen + 2;
							maxlen = 0;
							if(!i) maxlen = 16;
						}
					}
				}
				if(!i) {
					entery = ypos + y;
					enterx = xpos;
				}
				xpos += maxlen + 2;

				color_set(0, 0);
				move(ypos + 1, firstx);
				for(x = firstx; x < xpos - 2; x++) addch(ACS_S3);
			}
		}
	}

	if(evaluating) {
		if(eval_phase == EP_ANIMATEINPUT) {
			entery = 5;
			enterx = 33;
		} else {
			entery = 9;
			enterx = 24;
		}
	}

	move(entery, enterx);
	inpbuf[inppos] = 0;
	if(wcstombs(buf, inpbuf, sizeof(buf)) > 0) {
		addstr(buf);
	}

	/*if(evaluating && eval_phase == EP_VOTE) {
		move(((votecursor == nsuggest) ? 13 : 12) + votecursor, 2);
	}*/
}

struct client *findclient(char *name, int create) {
	int i;

	for(i = 0; i < MAXCLIENTS; i++) {
		if(client[i].status != CS_NO
		&& !strcmp(client[i].name, name)) {
			return &client[i];
		}
	}

	if(!create) return 0;

	for(i = 0; i < MAXCLIENTS; i++) {
		if(client[i].status == CS_NO) {
			client[i].status = CS_LOBBY;
			strncpy(client[i].name, name, sizeof(client[i].name));
			return &client[i];
		}
	}

	exit(2);
}

int contribcmp(const void *a, const void *b) {
	return wcscoll(
		((const struct contrib *) a)->word,
		((const struct contrib *) b)->word);
}

int suggestcmp(const void *a, const void *b) {
	return wcscoll(
		((const struct suggest *) a)->word,
		((const struct suggest *) b)->word);
}

void handlemsg(char *m) {
	int i, j, k;
	int i1;
	char *s1, *s2;
	struct client *c;
	wchar_t temp[17];
	struct contrib *ptr;
	struct itimerval timr;
	struct suggest *sugg;

	if(1 == sscanf(m, "in lobby %as", &s1)) {
		findclient(s1, 1)->status = CS_LOBBY;
		free(s1);
	} else if(1 == sscanf(m, "playing %as", &s1)) {
		findclient(s1, 1)->status = CS_PLAYING;
		free(s1);
	} else if(1 == sscanf(m, "disconnect %as", &s1)) {
		c = findclient(s1, 0);
		if(c) c->status = CS_NO;
		free(s1);
	} else if(2 == sscanf(m, "game %d %as", &i1, &s1)) {
		status = CS_PLAYING;
		evaluating = 0;
		inppos = 0;
		game.nsol = i1;
		game.nsol_left = i1;
		j = 0;
		setlocale(LC_CTYPE, "en_US.utf8");
		for(i = 0; i < 16; i++) {
			k = mbtowc(&game.box[i], &s1[j], strlen(s1) - j);
			if(k > 0) j += k;
		}
		setlocale(LC_CTYPE, "");
		for(i = 0; i < MAXCLIENTS; i++) {
			if(client[i].status == CS_LOBBY) {
				client[i].status = CS_PLAYING;
			}
			if(client[i].status == CS_PLAYING) {
				for(j = 0; j < client[i].ncontrib; j++) {
					free(client[i].contrib[i].word);
				}
				free(client[i].contrib);
				client[i].contrib = 0;
				client[i].ncontrib = 0;
				client[i].ncontriballoc = 0;
			}
		}
		game.timer = gametime * 10;
		timr.it_interval.tv_sec = 0;
		timr.it_interval.tv_usec = 100000;
		timr.it_value.tv_sec = 0;
		timr.it_value.tv_usec = 100000;
		setitimer(ITIMER_REAL, &timr, 0);
		free(s1);
	} else if(!strcmp(m, "evaluating") && status != CS_LOBBY) {
		evaluating = 1;
		eval_phase = EP_ANIMATE;
	} else if(
			status != CS_LOBBY && (
			(2 == (i = sscanf(m, "contrib %as %a[^\n]", &s1, &s2))) ||
			(1 == (i = sscanf(m, "missed %a[^\n]", &s2))))) {

		temp[16] = 0;
		if(i == 2) {
			c = findclient(s1, 0);
		} else {
			s1 = 0;
			c = &client[MAXCLIENTS - 1];
		}
		setlocale(LC_CTYPE, "en_US.utf8");
		if(c && (mbstowcs(temp, s2, 16) > 0)) {
			if(c->ncontrib >= c->ncontriballoc) {
				j = c->ncontriballoc * 2 + 1;
				ptr = malloc(sizeof(struct contrib) * j);
				if(!ptr) {
					perror("malloc");
					exit(1);
				}
				memcpy(ptr, c->contrib, sizeof(struct contrib) * c->ncontrib);
				free(c->contrib);
				c->contrib = ptr;
				c->ncontriballoc = j;
			}
			c->contrib[c->ncontrib].word = wcsdup(temp);
			c->contrib[c->ncontrib].flashthresh = gametime * 10 + 1;
			c->ncontrib++;
		}
		setlocale(LC_CTYPE, "");
		if(i == 1) {
			if (!(game.nsol_left - c->ncontrib))
				qsort(c->contrib, c->ncontrib, sizeof(struct contrib), contribcmp);
		} else {
			game.nsol_left--;
		}
		free(s1);
		free(s2);
	} else if(!strcmp(m, "timeout") && status != CS_LOBBY) {
		evaluating = 1;
		eval_phase = EP_ANIMATE;
		gameovertxt = "Game over";
	} else if(!strcmp(m, "solved") && status != CS_LOBBY) {
		evaluating = 1;
		eval_phase = EP_ANIMATE;
		gameovertxt = "Puzzle solved";
	} else if(1 == sscanf(m, "animate %as", &s1) && status != CS_LOBBY) {
		if(strlen(s1) <= 16) {
			animate_len = strlen(s1);
			animate_pos = 0;
			for(i = 0; i < animate_len; i++) {
				animate_path[i] = s1[i] - 'a';
			}
			animate_flag = 1;
		}
		free(s1);
	} else if(1 == sscanf(m, "suggest %as", &s1) && status != CS_LOBBY) {
		setlocale(LC_CTYPE, "en_US.utf8");
		if(mbstowcs(temp, s1, 17) > 0
		&& wcslen(temp) > 2
		&& wcslen(temp) <= 16) {
			if(nsuggest >= nsuggestalloc) {
				i = nsuggestalloc * 2 + 1;
				sugg = malloc(sizeof(struct suggest) * i);
				if(!sugg) {
					perror("malloc");
					exit(1);
				}
				memcpy(sugg, suggest, sizeof(struct suggest) * nsuggest);
				free(suggest);
				suggest = sugg;
				nsuggestalloc = i;
			}
			wcscpy(suggest[nsuggest].word, temp);
			suggest[nsuggest].vote = 0;
			nsuggest++;
		}
		free(s1);
		setlocale(LC_CTYPE, "");
	} else if(1 == sscanf(m, "gametime %d", &i1)) {
		gametime = i1;
	} else if(!strcmp(m, "lastsuggest") && status != CS_LOBBY) {
		qsort(suggest, nsuggest, sizeof(struct suggest), suggestcmp);
	} else if(1 == sscanf(m, "timer %d", &i1)) {
		game.timer = i1 * 10;
	}
}

void return_to_lobby() {
	int i, j;

	for(i = 0; i < MAXCLIENTS; i++) {
		for(j = 0; j < client[i].ncontrib; j++) {
			free(client[i].contrib[j].word);
		}
		free(client[i].contrib);
		client[i].contrib = 0;
		client[i].ncontrib = 0;
		client[i].ncontriballoc = 0;
		free(suggest);
		suggest = 0;
		nsuggest = 0;
		nsuggestalloc = 0;
	}
	status = CS_LOBBY;
	msg("in lobby");
}

void try_guess(wchar_t *w) {
	char buf[1024];
	int i, j, t;

	for(i = 0; i < MAXCLIENTS; i++) {
		if(client[i].status == CS_PLAYING) {
			for(j = 0; j < client[i].ncontrib; j++) {
				if(!wcscmp(client[i].contrib[j].word, w)) {
					t = game.timer - 8;
					if(t < 0) t = 0;
					client[i].contrib[j].flashthresh = t;
					return;
				}
			}
		}
	}

	setlocale(LC_CTYPE, "en_US.utf8");
	if(wcstombs(buf, w, sizeof(buf)) > 0) {
		msg("guess %s", buf);
	}
	setlocale(LC_CTYPE, "");
}

mbstate_t inpkeystate;

void handlekey(int key) {
	wchar_t temp;
	char ch;
	char buf[1024];
	int i;

	if(key == 12) {
		redrawwin(stdscr);
		return;
	}

	if(status == CS_LOBBY) {
		if(key == ' ') {
			msg("request new");
		}
	} else if(status == CS_PLAYING) {
		if(evaluating && inppos && (key == 10 || key == 13 || key == KEY_ENTER) && eval_phase != EP_ANIMATEINPUT) {
			//hantera nr man skriver ord och slr enter just nr tiden tar slut
			inppos = 0;
		} else if(!evaluating || (evaluating && eval_phase == EP_ANIMATEINPUT)) {
			if(key == 8 || key == KEY_BACKSPACE) {
				if(inppos) inppos--;
			} else if(key == 10 || key == 13 || key == KEY_ENTER) {
				inpbuf[inppos] = 0;
				if(evaluating) {
					setlocale(LC_CTYPE, "en_US.utf8");
					if(wcstombs(buf, inpbuf, sizeof(buf)) >= 0) {
						msg("animrequest %s", buf);
						eval_phase = EP_ANIMATE;
					}
					setlocale(LC_CTYPE, "");
				} else {
					try_guess(inpbuf);
				}
				inppos = 0;
			} else if(key <= ' ' || key == 127 || key > 255) {
				/* nop */
			} else {
				ch = key;
				if((inppos < 16)
				&& (1 == mbrtowc(&temp, &ch, 1, &inpkeystate))) {
					inpbuf[inppos++] = temp;
				}
			}
		} else {
			switch(eval_phase) {
				case EP_ANIMATE:
					if(key == 'a') {
						eval_phase = EP_ANIMATEINPUT;
						inppos = 0;
					} else if(key == 10 || key == 13 || key == KEY_ENTER) {
						animate_flag = 0;
						if(nsuggest) {
							eval_phase = EP_VOTE;
							votecursor = 0;
						} else {
							return_to_lobby();
						}
					}
					break;
				case EP_VOTE:
					if(key == 10 || key == 13 || key == KEY_ENTER || key == ' ') {
						if(votecursor == nsuggest) {
							setlocale(LC_CTYPE, "en_US.utf8");
							for(i = 0; i < nsuggest; i++) {
								if(suggest[i].vote) {
									if(wcstombs(buf, suggest[i].word, sizeof(buf)) > 0) {
										msg("vote %s", buf);
									}
								}
							}
							setlocale(LC_CTYPE, "");
							return_to_lobby();
						} else {
							suggest[votecursor].vote ^= 1;
						}
					} else if(key == KEY_UP || key == 'k') {
						if(votecursor) votecursor--;
					} else if(key == KEY_DOWN || key == 'j') {
						if(votecursor < nsuggest) votecursor++;
					}
					break;
			}
		}
	}
}

void newsize() {
	struct winsize w;

	if(ioctl(0, TIOCGWINSZ, &w) >= 0) {
		dispwidth = w.ws_col;
		dispheight = w.ws_row;
		msg("FOUND size %d %d", dispwidth, dispheight);
	}
	endwin();
	redrawwin(stdscr);
	initscr();
}

ssize_t forceread(int fd, char *ptr, size_t len) {
	ssize_t res;
	size_t left = len;

	do {
		res = read(fd, ptr, left);
		if(res == 0) return 0;
		if(res < 0) return res;
		left -= res;
		ptr += res;
	} while(left);
	return len;
}

int main(int argc, char **argv) {
	char *server;
	struct hostent *host;
	struct sockaddr_in sockaddr;
	int running = 1;
	char buf[1024];
	fd_set fds;
	ssize_t size;
	char *exitmsg = 0;
	uint32_t addr;

	memset(client, 0, sizeof(client));
	memset(&game, 0, sizeof(game));

	setlocale(LC_ALL, "");

	if(argc != 3) {
		fprintf(stderr, "Usage: %s nick server\n", argv[0]);
		exit(1);
	}

	nick = strndup(argv[1], 10);
	server = argv[2];

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if(sock < 0) {
		perror("socket");
		exit(1);
	}

	printf("Looking up %s\n", server);
	host = gethostbyname2(server, AF_INET);

	if(!host) {
		perror("gethostbyname");
		exit(1);
	}

	sockaddr.sin_family = AF_INET;
	sockaddr.sin_port = htons(1789);
	memcpy(&sockaddr.sin_addr, host->h_addr_list[0], sizeof(sockaddr.sin_addr));

	addr = ntohl(sockaddr.sin_addr.s_addr);
	printf("Connecting to %d.%d.%d.%d\n",
		(addr >> 24) & 255,
		(addr >> 16) & 255,
		(addr >>  8) & 255,
		(addr >>  0) & 255);

	if(connect(sock, (struct sockaddr *) &sockaddr, sizeof(sockaddr))) {
		perror("connect");
		exit(1);
	}

	buf[MSGLENGTH] = 0;
	if(forceread(sock, buf, MSGLENGTH) != MSGLENGTH) {
		fprintf(stderr, "Communication error\n");
		exit(1);
	}
	if(1 != sscanf(buf, "welcome %[^\n]", serverid)) {
		fprintf(stderr, "Communication error\n", buf);
		exit(1);
	}

	printf("Connected to %s\n", serverid);

	msg("protocol %d", PROTVER);

	if(forceread(sock, buf, MSGLENGTH) != MSGLENGTH) {
		fprintf(stderr, "Communication error\n");
		exit(1);
	}
	if(strcmp(buf, "proto ok")) {
		fprintf(stderr, "Please update your client at http://kryo.se/distboggle\n");
		exit(1);
	}

	msg("nick %s", nick);

	if(forceread(sock, buf, MSGLENGTH) != MSGLENGTH) {
		fprintf(stderr, "Communication error\n");
		exit(1);
	}
	if(strcmp(buf, "nick ok")) {
		fprintf(stderr, "Nick taken or invalid\n");
		exit(1);
	}

	client[0].status = CS_LOBBY;
	strcpy(client[0].name, nick);
	client[0].contrib = 0;
	client[0].ncontrib = 0;
	client[0].ncontriballoc = 0;

	client[MAXCLIENTS - 1].status = CS_MISSED;
	strcpy(client[MAXCLIENTS - 1].name, "you missed");
	client[MAXCLIENTS - 1].contrib = 0;
	client[MAXCLIENTS - 1].ncontrib = 0;
	client[MAXCLIENTS - 1].ncontriballoc = 0;

	signal(SIGINT, sighandler);
	signal(SIGALRM, sighandler);
	signal(SIGWINCH, sighandler);

	initscr();
	start_color();
	init_pair(1, COLOR_CYAN, COLOR_BLACK);
	init_pair(2, COLOR_YELLOW, COLOR_BLACK);
	init_pair(3, COLOR_BLACK, COLOR_GREEN);
	cbreak();
	noecho();
	keypad(stdscr, 1);

	newsize();

	while(running) {
		erase();
		mvaddstr(0, 0, "distBoggle by KRYO");
		snprintf(buf, sizeof(buf), "Server: %s", serverid);
		mvaddstr(0, 24, buf);
		snprintf(buf, sizeof(buf), "Nick: %s", nick);
		mvaddstr(0, 50, buf);
		switch(status) {
			case CS_LOBBY:
				disp_lobby();
				break;
			case CS_PLAYING:
				disp_playing();
				break;
		}
		refresh();

		FD_ZERO(&fds);
		FD_SET(0, &fds);
		FD_SET(sock, &fds);
		if(select(sock + 1, &fds, 0, 0, 0) >= 0) {
			if(FD_ISSET(0, &fds)) {
				handlekey(getch());
			}
			if(FD_ISSET(sock, &fds)) {
				size = forceread(sock, buf, MSGLENGTH);
				if(size == 0) {
					running = 0;
					exitmsg = "Connection to server lost";
				} else if(size != MSGLENGTH) {
					running = 0;
					exitmsg = "Communication error";
				} else {
					handlemsg(buf);
				}
			}
		}

		if(got_ctrl_c) {
			running = 0;
		}
		if(got_winch) {
			got_winch = 0;
			newsize();
		}
	}

	endwin();
	if(exitmsg) printf("%s\n", exitmsg);
	return 0;
}
